{# TEMPLATE VAR SETTINGS #}
{%- set url_root = pathto('', 1) %}
{%- if url_root == '#' %}{% set url_root = '' %}{% endif %}
{%- if not embedded and docstitle %}
  {%- set titlesuffix = " &mdash; "|safe + docstitle|e %}
{%- else %}
  {%- set titlesuffix = "" %}
{%- endif %}

<!DOCTYPE html>
<!--[if IE 8]><html class="no-js lt-ie9" lang="en" > <![endif]-->
<!--[if gt IE 8]><!--> <html class="no-js" lang="en" > <!--<![endif]-->
<head>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1.0">

  <script type="text/javascript" src="//www.redhat.com/dtm.js"></script>

  <!-- Google Tag Manager Data Layer -->
  <script>
   dataLayer = [];
  </script>
  <!-- End Google Tag Manager Data Layer -->




  {% block htmltitle %}
  <title>{{ title|striptags|e }}{{ titlesuffix }}</title>
  {% endblock %}

  {# FAVICON #}
  {% if favicon %}
    <link rel="shortcut icon" href="{{ pathto('_static/' + favicon, 1) }}"/>
  {% endif %}

  {# CSS #}
  <link href='https://fonts.googleapis.com/css?family=Lato:400,700|Roboto+Slab:400,700|Inconsolata:400,700' rel='stylesheet' type='text/css'>
  <link href='https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.0.3/css/font-awesome.min.css' rel='stylesheet' type='text/css'>

  {# JS #}
  {% if not embedded %}

    <script type="text/javascript">
      var DOCUMENTATION_OPTIONS = {
        URL_ROOT:'{{ url_root }}',
        VERSION:'{{ release|e }}',
        COLLAPSE_INDEX:false,
        FILE_SUFFIX:'{{ '' if no_search_suffix else file_suffix }}',
        HAS_SOURCE:  {{ has_source|lower }}
      };
    </script>
    {%- for scriptfile in script_files %}
      <script type="text/javascript" src="{{ pathto(scriptfile, 1) }}"></script>
    {%- endfor %}

    {% if use_opensearch %}
      <link rel="search" type="application/opensearchdescription+xml" title="{% trans docstitle=docstitle|e %}Search within {{ docstitle }}{% endtrans %}" href="{{ pathto('_static/opensearch.xml', 1) }}"/>
    {% endif %}

  {% endif %}

  {# RTD hosts these file themselves, so just load on non RTD builds #}
  {% if not READTHEDOCS %}
    <link rel="stylesheet" href="{{ pathto('_static/' + style, 1) }}" type="text/css" />
    <script type="text/javascript" src="_static/js/theme.js"></script>
  {% endif %}

  {% for cssfile in css_files %}
    <link rel="stylesheet" href="{{ pathto(cssfile, 1) }}" type="text/css" />
  {% endfor %}

  {%- block linktags %}
    {%- if hasdoc('about') %}
        <link rel="author" title="{{ _('About these documents') }}"
              href="{{ pathto('about') }}"/>
    {%- endif %}
    {%- if hasdoc('genindex') %}
        <link rel="index" title="{{ _('Index') }}"
              href="{{ pathto('genindex') }}"/>
    {%- endif %}
    {%- if hasdoc('search') %}
        <link rel="search" title="{{ _('Search') }}" href="{{ pathto('search') }}"/>
    {%- endif %}
    {%- if hasdoc('copyright') %}
        <link rel="copyright" title="{{ _('Copyright') }}" href="{{ pathto('copyright') }}"/>
    {%- endif %}
    <link rel="top" title="{{ docstitle|e }}" href="{{ pathto('index') }}"/>
    {%- if parents %}
        <link rel="up" title="{{ parents[-1].title|striptags|e }}" href="{{ parents[-1].link|e }}"/>
    {%- endif %}
    {%- if next %}
        <link rel="next" title="{{ next.title|striptags|e }}" href="{{ next.link|e }}"/>
    {%- endif %}
    {%- if prev %}
        <link rel="prev" title="{{ prev.title|striptags|e }}" href="{{ prev.link|e }}"/>
    {%- endif %}
  {%- endblock %}
  {%- block extrahead %} {% endblock %}

  <script src="//cdnjs.cloudflare.com/ajax/libs/modernizr/2.6.2/modernizr.min.js"></script>

  <style>
    .search-reset-start {
        color: #463E3F;
        float: right;
        position: relative;
        top: -25px;
        left: -10px;
        z-index: 10;
     }
     .search-reset-start:hover {
        cursor: pointer;
        color: #2980B9;
     }
     #search-box-id {
        padding-right: 25px;
     }
  </style>

  <link href='https://fonts.googleapis.com/css?family=Open+Sans:400,300'
    rel='stylesheet' type='text/css'>

</head>

<body class="wy-body-for-nav">

  <!-- Google Tag Manager -->
  <noscript><iframe src="//www.googletagmanager.com/ns.html?id=GTM-PSB293" height="0" width="0" style="display:none;visibility:hidden"></iframe></noscript>
  <script>(function(w,d,s,l,i){w[l]=w[l]||[];w[l].push({'gtm.start': new Date().getTime(),event:'gtm.js'});var f=d.getElementsByTagName(s)[0], j=d.createElement(s),dl=l!='dataLayer'?'&l='+l:'';j.async=true;j.src='//www.googletagmanager.com/gtm.js?id='+i+dl;f.parentNode.insertBefore(j,f); })(window,document,'script','dataLayer','GTM-PSB293');</script>
  <!-- End Google Tag Manager -->

    <div class="DocSite-globalNav ansibleNav">
        <ul>
            <li><a href="https://www.ansible.com/ansiblefest" target="_blank">AnsibleFest</a></li>
            <li><a href="https://www.ansible.com/tower" target="_blank">Products</a></li>
            <li><a href="https://www.ansible.com/community" target="_blank">Community</a></li>
            <li><a href="https://www.ansible.com/webinars-training" target="_blank">Webinars & Training</a></li>
            <li><a href="https://www.ansible.com/blog" target="_blank">Blog</a></li>
        </ul>
    </div>

  <a class="DocSite-nav" href="/">
    <img class="DocSiteNav-logo"
      src="{{ pathto('_static/', 1) }}images/logo_invert.png"
      alt="Ansible Logo">
    <div class="DocSiteNav-title">
      Documentation
    </div>
  </a>

  <div class="wy-grid-for-nav">

    {# SIDE NAV, TOGGLES ON MOBILE #}
    <nav data-toggle="wy-nav-shift" class="wy-nav-side">

      <div style="background-color:#5bbdbf;height=80px;margin:'auto auto auto auto'">
        <a class="DocSiteProduct-header DocSiteProduct-header--core" href="/">
            <div class="DocSiteProduct-productName">
                <div class="DocSiteProduct-logoText">
                    Ansible  
                    <div class="DocSiteProduct-CurrentVersion" align="right">
                      devel
                  </div>
                </div>
            </div>
        </a>
        

          <div class="DocSiteProduct-CheckVersionPara">For previous versions, see the  <a class="DocSiteProduct-versionheader" href="/#coreversionselect">documentation archive.</a></div>
      </div>

      <div id="menu-id" class="wy-menu wy-menu-vertical" data-spy="affix">
        {% set toctree = toctree(maxdepth=4, collapse=True) %}
        {% if toctree %}
            {{ toctree }}
        {% else %}
            {{ toc }}
        {% endif %}


    <div class="DocSite-sideNav ansibleNav">
       <ul class="">
       <li><a href="https://www.ansible.com/ansiblefest" target="_blank">AnsibleFest</a></li>
       <li><a href="https://www.ansible.com/tower" target="_blank">Products</a></li>
       <li><a href="https://www.ansible.com/community" target="_blank">Community</a></li>
       <li><a href="https://www.ansible.com/webinars-training" target="_blank">Webinars & Training</a></li>
       <li><a href="https://www.ansible.com/blog" target="_blank">Blog</a></li>
       </ul>
     </div>

    <!-- changeable widget -->
    <div id="sideBanner">
    <br/>
    <a href="https://www.ansible.com/docs-left?utm_source=docs">
      <img style="border-width:0px;" src="https://cdn2.hubspot.net/hubfs/330046/docs-graphics/ASB-docs-left-rail.png" />
    </a>
    </div>


      </div>
      &nbsp;
    </nav>

    <section data-toggle="wy-nav-shift" class="wy-nav-content-wrap">

      {# MOBILE NAV, TRIGGLES SIDE NAV ON TOGGLE #}
      <nav class="wy-nav-top">
        <i data-toggle="wy-nav-top" class="icon icon-reorder"></i>
        <a href="/">{{ project }}</a>
      </nav>


      {# PAGE CONTENT #}
      <div class="wy-nav-content">
        <div class="rst-content">

          <!-- Banner ads
          <div class="DocSiteBanner">
            <a class="DocSiteBanner-imgWrapper"
                href="https://www.ansible.com/docs-top?utm_source=docs">
                <img src="https://cdn2.hubspot.net/hubfs/330046/docs-graphics/ASB-docs-top-left.png">
            </a>
            <a class="DocSiteBanner-imgWrapper"
                href="https://www.ansible.com/docs-top?utm_source=docs">
                <img src="https://cdn2.hubspot.net/hubfs/330046/docs-graphics/ASB-docs-top-right.png">
            </a>
        </div> -->

          {% include "breadcrumbs.html" %}
          <div id="page-content">
          {% block body %}{% endblock %}
          </div>
          <div id="search-results"><gcse:searchresults-only></div>
          {% include "footer.html" %}
        </div>
      </div>

    </section>

  </div>
  {% include "versions.html" %}

<!-- begin analytics -->
<script type="text/javascript">
var _hsq = _hsq || [];
_hsq.push(["setContentType", "standard-page"]);
        (function(d,s,i,r) {
        if (d.getElementById(i)){return;}
        var n = d.createElement(s),e = document.getElementsByTagName(s)[0];
        n.id=i;n.src = '//js.hs-analytics.net/analytics/'+(Math.ceil(new Date()/r)*r)+'/330046.js';
        e.parentNode.insertBefore(n, e);
        })(document, "script", "hs-analytics",300000);
</script>
<!-- end analytics -->
<script type="text/javascript">
if (("undefined" !== typeof _satellite) && ("function" === typeof _satellite.pageBottom)) {
_satellite.pageBottom();
}
</script>
</body>
</html>
