no ip access-list extended test
ip access-list extended test
 permit ip host 192.0.2.1 any log
 permit ip host 192.0.2.2 any log
 permit ip host 192.0.2.3 any log
 permit ip host 192.0.2.4 any log
 permit ip host 192.0.2.5 any log
